"""
Voronoi tesslation of 2-d point sets.

Adapted from https://gist.github.com/pv/8036995

"""

import numpy as np
from scipy.spatial import Voronoi

__author__ = "Serge Rey <sjsrey@gmail.com>"

__all__ = ["voronoi_frames"]


def voronoi(points, radius=None):
    """Determine finite Voronoi diagram for a 2-d point set.
    See also ``voronoi_regions()``.

    Parameters
    ----------
    points : array_like
        An nx2 array of points.
    radius : float (optional)
        The distance to 'points at infinity'. Default is ``None.``

    Returns
    -------
    vor : tuple
        A two-element tuple consisting of a list and an array. Each element of
        the list contains the sequence of the indices of Voronoi vertices
        composing a Voronoi polygon (region), whereas the array contains
        the Voronoi vertex coordinates.
    
    Examples
    --------
    
    >>> points = [(10.2, 5.1), (4.7, 2.2), (5.3, 5.7), (2.7, 5.3)]
    >>> regions, coordinates = voronoi(points)
    >>> regions
    [[1, 3, 2], [4, 5, 1, 0], [0, 1, 7, 6], [9, 0, 8]]
    
    >>> coordinates
    array([[  4.21783296,   4.08408578],
           [  7.51956025,   3.51807539],
           [  9.4642193 ,  19.3994576 ],
           [ 14.98210684, -10.63503022],
           [ -9.22691341,  -4.58994414],
           [ 14.98210684, -10.63503022],
           [  1.78491801,  19.89803294],
           [  9.4642193 ,  19.3994576 ],
           [  1.78491801,  19.89803294],
           [ -9.22691341,  -4.58994414]])
    
    """

    vor = voronoi_regions(Voronoi(points), radius=radius)

    return vor


def voronoi_regions(vor, radius=None):
    """Finite voronoi regions for a 2-d point set. See also ``voronoi()``.

    Parameters
    ----------
    vor : scipy.spatial.Voronoi
        A planar Voronoi diagram.
    radius : float (optional)
        Distance to 'points at infinity'. Default is ``None.``
    
    Returns
    -------
    regions_vertices : tuple
        A two-element tuple consisting of a list of finite voronoi regions
        and an array Voronoi vertex coordinates.
    
    """

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max() * 2

    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            new_regions.append(vertices)
            continue

        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                continue

            t = vor.points[p2] - vor.points[p1]
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        new_regions.append(new_region.tolist())

    regions_vertices = new_regions, np.asarray(new_vertices)

    return regions_vertices


def as_dataframes(regions, vertices, points):
    """Helper function to store finite Voronoi regions and
    originator points as ``geopandas`` (or ``pandas``) dataframes.

    Parameters
    ----------
    regions : list
        Each element of the list contains sequence of the indexes of
        voronoi vertices composing a vornoi polygon (region).
    vertices : array_like
        The coordinates of the vornoi vertices.
    points : array_like
        The originator points.

    Returns
    -------
    region_df : geopandas.GeoDataFrame
        Finite Voronoi polygons as geometries.
    points_df : geopandas.GeoDataFrame
        Originator points as geometries.
    
    Raises
    ------
    ImportError
        Raised when ``geopandas`` is not available.
    ImportError
        Raised when ``shapely`` is not available.
    
    """

    try:
        import geopandas as gpd
    except ImportError:
        gpd = None

    try:
        from shapely.geometry import Polygon, Point
    except ImportError:
        from .shapes import Polygon, Point

    if gpd is not None:
        region_df = gpd.GeoDataFrame()
        region_df["geometry"] = [Polygon(vertices[region]) for region in regions]

        point_df = gpd.GeoDataFrame()
        point_df["geometry"] = gpd.GeoSeries(Point(pnt) for pnt in points)
    else:
        import pandas as pd

        region_df = pd.DataFrame()
        region_df["geometry"] = [
            Polygon(vertices[region].tolist()) for region in regions
        ]
        point_df = pd.DataFrame()
        point_df["geometry"] = [Point(pnt) for pnt in points]

    return region_df, point_df


def voronoi_frames(points, radius=None, clip="extent"):
    """Composite helper to return Voronoi regions and
    generator points as individual dataframes.

    Parameters
    ----------
    points : array_like
        The originator points.
    radius : float
        The distance to 'points at infinity' used in building voronoi cells.
        Default is ``None``.
    clip : {str, shapely.geometry.Polygon}
        An overloaded option about how to clip the voronoi cells.
        Default is ``'extent'``. Options are as follows.
        
        * ``'none'``/``None`` -- No clip is applied. Voronoi cells may be arbitrarily larger that the source map. Note that this may lead to cells that are many orders of magnitude larger in extent than the original map. Not recommended.
        * ``'bbox'``/``'extent'``/``'bounding box'`` -- Clip the voronoi cells to the bounding box of the input points.
        * ``'chull``/``'convex hull'`` -- Clip the voronoi cells to the convex hull of the input points.
        * ``'ashape'``/``'ahull'`` -- Clip the voronoi cells to the tightest hull that contains all points (e.g. the smallest alphashape, using ``libpysal.cg.alpha_shape_auto``).
        * Polygon -- Clip to an arbitrary Polygon.

    Returns
    -------
    reg_vtx : tuple
        Two ``geopandas.GeoDataFrame`` (or ``pandas.DataFrame`` if ``geopandas``
        is unavailable) objects--``(region_df, points_df)``--of finite 
        Voronoi polygons and the originator points as geometries.

    Notes
    -----

    If ``geopandas`` is not available the return types will be
    ``pandas.DataFrame`` objects, each with a geometry column populated
    with PySAL shapes. If ``geopandas`` is available, return types are
    ``pandas.GeoDataFrame`` objects with a geometry column populated
    with shapely geometry types.

    Examples
    --------
    
    >>> points = [(10.2, 5.1), (4.7, 2.2), (5.3, 5.7), (2.7, 5.3)]
    >>> regions_df, points_df = voronoi_frames(points)
    >>> regions_df.shape
    (4, 1)
    
    >>> regions_df.shape == points_df.shape
    True

    """

    regions, vertices = voronoi(points, radius=radius)
    regions, vertices = as_dataframes(regions, vertices, points)
    if clip:
        regions = clip_voronoi_frames_to_extent(regions, vertices, clip=clip)

    reg_vtx = regions, vertices
    return reg_vtx


def clip_voronoi_frames_to_extent(regions, vertices, clip="extent"):
    """Generate a geopandas.GeoDataFrame of Voronoi cells clipped to
    a specified extent.
    
    Parameters
    ----------
    regions : geopandas.GeoDataFrame
        A (geo)dataframe containing voronoi cells to clip.
    vertices : geopandas.GeoDataFrame
        A (geo)dataframe containing vertices used to build voronoi cells.
    clip : str, shapely.geometry.Polygon
        An overloaded option about how to clip the voronoi cells.
        The options are:
          - 'none'/None: No clip is applied. Voronoi cells may be arbitrarily
            larger that the source map. Note that this may lead to cells that
            are many orders of magnitude larger in extent than
            the original map. Not recommended.
          - 'bbox'/'extent'/'bounding box': Clip the voronoi cells to the
            bounding box of the input points.
          - 'chull'/'convex hull': Clip the voronoi cells to the
            convex hull of the input points.
          - 'ashape'/'ahull': Clip the voronoi cells to the tightest hull that
            contains all points (e.g. the smallest alphashape,
            using ``libpysal.cg.alpha_shape_auto``).
          - Polygon: Clip to an arbitrary Polygon.
    
    Returns
    -------
    clipped_regions : geopandas.GeoDataFrame
        A ``geopandas.GeoDataFrame`` of clipped voronoi regions.
    
    Raises
    ------
    ImportError
        Raised when ``shapely`` is not available.
    ImportError
        Raised when ``geopandas`` is not available.
    ValueError
        Raised when in invalid value for ``clip`` is passed in.
    
    """
    try:
        from shapely.geometry import Polygon
    except ImportError:
        raise ImportError("Shapely is required to clip voronoi regions.")
    try:
        import geopandas
    except ImportError:
        raise ImportError("Geopandas is required to clip voronoi regions.")

    if isinstance(clip, Polygon):
        clipper = geopandas.GeoDataFrame(geometry=[clip])
    elif clip is None:
        return regions
    elif clip.lower() == "none":
        return regions
    elif clip.lower() in ("bounds", "bounding box", "bbox", "extent"):
        min_x, min_y, max_x, max_y = vertices.total_bounds
        bounding_poly = Polygon(
            [
                (min_x, min_y),
                (min_x, max_y),
                (max_x, max_y),
                (max_x, min_y),
                (min_x, min_y),
            ]
        )
        clipper = geopandas.GeoDataFrame(geometry=[bounding_poly])
    elif clip.lower() in ("chull", "convex hull", "convex_hull"):
        clipper = geopandas.GeoDataFrame(
            geometry=[vertices.geometry.unary_union.convex_hull]
        )
    elif clip.lower() in (
        "ahull",
        "alpha hull",
        "alpha_hull",
        "ashape",
        "alpha shape",
        "alpha_shape",
    ):
        from .alpha_shapes import alpha_shape_auto
        from ..weights.distance import get_points_array

        coordinates = get_points_array(vertices.geometry)
        clipper = geopandas.GeoDataFrame(geometry=[alpha_shape_auto(coordinates)])
    else:
        raise ValueError(
            "Clip type '{}' not understood. Try one "
            " of the supported options: [None, 'extent', "
            "'chull', 'ahull'].".format(clip)
        )
    clipped_regions = geopandas.overlay(regions, clipper, how="intersection")
    return clipped_regions



__all__ = ["KNN", "Kernel", "DistanceBand"]
__author__ = "Sergio J. Rey <srey@asu.edu>, Levi John Wolf <levi.john.wolf@gmail.com>"


from ..cg.kdtree import KDTree
from .weights import W, WSP
from .util import (
    isKDTree,
    get_ids,
    get_points_array_from_shapefile,
    get_points_array,
    WSP2W,
)
import copy
from warnings import warn as Warn
from scipy.spatial import distance_matrix
import scipy.sparse as sp
import numpy as np


def knnW(data, k=2, p=2, ids=None, radius=None, distance_metric="euclidean"):
    """
    This is deprecated. Use the pysal.weights.KNN class instead.
    """
    # Warn('This function is deprecated. Please use pysal.weights.KNN', UserWarning)
    return KNN(data, k=k, p=p, ids=ids, radius=radius, distance_metric=distance_metric)


class KNN(W):
    """
    Creates nearest neighbor weights matrix based on k nearest
    neighbors.

    Parameters
    ----------
    kdtree      : object
                  PySAL KDTree or ArcKDTree where KDtree.data is array (n,k)
                  n observations on k characteristics used to measure
                  distances between the n objects
    k           : int
                  number of nearest neighbors
    p           : float
                  Minkowski p-norm distance metric parameter:
                  1<=p<=infinity
                  2: Euclidean distance
                  1: Manhattan distance
                  Ignored if the KDTree is an ArcKDTree
    ids         : list
                  identifiers to attach to each observation

    Returns
    -------

    w         : W
                instance
                Weights object with binary weights

    Examples
    --------
    >>> import libpysal
    >>> import numpy as np
    >>> points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
    >>> kd = libpysal.cg.KDTree(np.array(points))
    >>> wnn2 = libpysal.weights.KNN(kd, 2)
    >>> [1,3] == wnn2.neighbors[0]
    True
    >>> wnn2 = KNN(kd,2)
    >>> wnn2[0]
    {1: 1.0, 3: 1.0}
    >>> wnn2[1]
    {0: 1.0, 3: 1.0}

    now with 1 rather than 0 offset

    >>> wnn2 = libpysal.weights.KNN(kd, 2, ids=range(1,7))
    >>> wnn2[1]
    {2: 1.0, 4: 1.0}
    >>> wnn2[2]
    {1: 1.0, 4: 1.0}
    >>> 0 in wnn2.neighbors
    False

    Notes
    -----

    Ties between neighbors of equal distance are arbitrarily broken. 

    Further, if many points occupy the same spatial location (i.e. observations are 
    coincident), then you may need to increase k for those observations to 
    acquire neighbors at different spatial locations. For example, if five
    points are coincident, then their four nearest neighbors will all
    occupy the same spatial location; only the fifth nearest neighbor will
    result in those coincident points becoming connected to the graph as a
    whole. 

    Solutions to this problem include jittering the points (by adding
    a small random value to each observation's location) or by adding 
    higher-k neighbors only to the coincident points, using the
    weights.w_sets.w_union() function. 

    See Also
    --------
    :class:`libpysal.weights.weights.W`
    """

    def __init__(
        self,
        data,
        k=2,
        p=2,
        ids=None,
        radius=None,
        distance_metric="euclidean",
        **kwargs
    ):
        if radius is not None:
            distance_metric = "arc"
        if isKDTree(data):
            self.kdtree = data
            self.data = self.kdtree.data
        else:
            self.kdtree = KDTree(data, radius=radius, distance_metric=distance_metric)
            self.data = self.kdtree.data
        self.k = k
        self.p = p

        # these are both n x k+1
        distances, indices = self.kdtree.query(self.data, k=k + 1, p=p)
        full_indices = np.arange(self.kdtree.n)

        # if an element in the indices matrix is equal to the corresponding
        # index for that row, we want to mask that site from its neighbors
        not_self_mask = indices != full_indices.reshape(-1, 1)
        # if there are *too many duplicates per site*, then we may get some
        # rows where the site index is not in the set of k+1 neighbors
        # So, we need to know where these sites are
        has_one_too_many = not_self_mask.sum(axis=1) == (k + 1)
        # if a site has k+1 neighbors, drop its k+1th neighbor
        not_self_mask[has_one_too_many, -1] &= False
        not_self_indices = indices[not_self_mask].reshape(self.kdtree.n, -1)

        to_weight = not_self_indices
        if ids is None:
            ids = list(full_indices)
            named_indices = not_self_indices
        else:
            named_indices = np.asarray(ids)[not_self_indices]
        neighbors = {idx: list(indices) for idx, indices in zip(ids, named_indices)}

        W.__init__(self, neighbors, id_order=ids, **kwargs)

    @classmethod
    def from_shapefile(cls, filepath, *args, **kwargs):
        """
        Nearest neighbor weights from a shapefile.

        Parameters
        ----------

        data       : string
                     shapefile containing attribute data.
        k          : int
                     number of nearest neighbors
        p          : float
                     Minkowski p-norm distance metric parameter:
                     1<=p<=infinity
                     2: Euclidean distance
                     1: Manhattan distance
        ids        : list
                     identifiers to attach to each observation
        radius     : float
                     If supplied arc_distances will be calculated
                     based on the given radius. p will be ignored.

        Returns
        -------

        w         : KNN
                    instance; Weights object with binary weights.

        Examples
        --------

        Polygon shapefile
        >>> import libpysal
        >>> from libpysal.weights import KNN
        >>> wc=KNN.from_shapefile(libpysal.examples.get_path("columbus.shp"))
        >>> "%.4f"%wc.pct_nonzero
        '4.0816'
        >>> set([2,1]) == set(wc.neighbors[0])
        True
        >>> wc3=KNN.from_shapefile(libpysal.examples.get_path("columbus.shp"),k=3)
        >>> set(wc3.neighbors[0]) == set([2,1,3])
        True
        >>> set(wc3.neighbors[2]) == set([4,3,0])
        True


        Point shapefile

        >>> w=KNN.from_shapefile(libpysal.examples.get_path("juvenile.shp"))
        >>> w.pct_nonzero
        1.1904761904761905
        >>> w1=KNN.from_shapefile(libpysal.examples.get_path("juvenile.shp"),k=1)
        >>> "%.3f"%w1.pct_nonzero
        '0.595'

        Notes
        -----

        Ties between neighbors of equal distance are arbitrarily broken.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        return cls(get_points_array_from_shapefile(filepath), *args, **kwargs)

    @classmethod
    def from_array(cls, array, *args, **kwargs):
        """
        Creates nearest neighbor weights matrix based on k nearest
        neighbors.

        Parameters
        ----------
        array       : np.ndarray
                      (n, k) array representing n observations on
                      k characteristics used to measure distances
                      between the n objects
        **kwargs    : keyword arguments, see Rook

        Returns
        -------
        w         : W
                    instance
                    Weights object with binary weights

        Examples
        --------
        >>> from libpysal.weights import KNN
        >>> points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
        >>> wnn2 = KNN.from_array(points, 2)
        >>> [1,3] == wnn2.neighbors[0]
        True
        >>> wnn2 = KNN.from_array(points,2)
        >>> wnn2[0]
        {1: 1.0, 3: 1.0}
        >>> wnn2[1]
        {0: 1.0, 3: 1.0}

        now with 1 rather than 0 offset

        >>> wnn2 = KNN.from_array(points, 2, ids=range(1,7))
        >>> wnn2[1]
        {2: 1.0, 4: 1.0}
        >>> wnn2[2]
        {1: 1.0, 4: 1.0}
        >>> 0 in wnn2.neighbors
        False

        Notes
        -----

        Ties between neighbors of equal distance are arbitrarily broken.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        return cls(array, *args, **kwargs)

    @classmethod
    def from_dataframe(cls, df, geom_col=None, ids=None, *args, **kwargs):
        """
        Make KNN weights from a dataframe.

        Parameters
        ----------
        df      :   pandas.dataframe
                    a dataframe with a geometry column that can be used to
                    construct a W object
        geom_col :  string
                    the name of the column in `df` that contains the
                    geometries. Defaults to active geometry column.
        ids     :   string or iterable
                    if string, the column name of the indices from the dataframe
                    if iterable, a list of ids to use for the W
                    if None, df.index is used.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        if geom_col is None:
            geom_col = df.geometry.name
        pts = get_points_array(df[geom_col])
        if ids is None:
            ids = df.index.tolist()
        elif isinstance(ids, str):
            ids = df[ids].tolist()
        return cls(pts, *args, ids=ids, **kwargs)

    def reweight(self, k=None, p=None, new_data=None, new_ids=None, inplace=True):
        """
        Redo K-Nearest Neighbor weights construction using given parameters

        Parameters
        ----------
        new_data    : np.ndarray
                      an array containing additional data to use in the KNN
                      weight
        new_ids     : list
                      a list aligned with new_data that provides the ids for
                      each new observation
        inplace     : bool
                      a flag denoting whether to modify the KNN object
                      in place or to return a new KNN object
        k           : int
                      number of nearest neighbors
        p           : float
                      Minkowski p-norm distance metric parameter:
                      1<=p<=infinity
                      2: Euclidean distance
                      1: Manhattan distance
                      Ignored if the KDTree is an ArcKDTree

        Returns
        -------
        A copy of the object using the new parameterization, or None if the
        object is reweighted in place.
        """

        if new_data is not None:
            new_data = np.asarray(new_data).reshape(-1, 2)
            data = np.vstack((self.data, new_data)).reshape(-1, 2)
            if new_ids is not None:
                ids = copy.deepcopy(self.id_order)
                ids.extend(list(new_ids))
            else:
                ids = list(range(data.shape[0]))
        elif (new_data is None) and (new_ids is None):
            # If not, we can use the same kdtree we have
            data = self.kdtree
            ids = self.id_order
        elif (new_data is None) and (new_ids is not None):
            Warn("Remapping ids must be done using w.remap_ids")
        if k is None:
            k = self.k
        if p is None:
            p = self.p
        if inplace:
            self._reset()
            self.__init__(data, ids=ids, k=k, p=p)
        else:
            return KNN(data, ids=ids, k=k, p=p)


class Kernel(W):
    """
    Spatial weights based on kernel functions.

    Parameters
    ----------

    data        : array
                  (n,k) or KDTree where KDtree.data is array (n,k)
                  n observations on k characteristics used to measure
                  distances between the n objects
    bandwidth   : float
                  or array-like (optional)
                  the bandwidth :math:`h_i` for the kernel.
    fixed       : binary
                  If true then :math:`h_i=h \\forall i`. If false then
                  bandwidth is adaptive across observations.
    k           : int
                  the number of nearest neighbors to use for determining
                  bandwidth. For fixed bandwidth, :math:`h_i=max(dknn) \\forall i`
                  where :math:`dknn` is a vector of k-nearest neighbor
                  distances (the distance to the kth nearest neighbor for each
                  observation).  For adaptive bandwidths, :math:`h_i=dknn_i`
    diagonal    : boolean
                  If true, set diagonal weights = 1.0, if false (default),
                  diagonals weights are set to value according to kernel
                  function.
    function    : {'triangular','uniform','quadratic','quartic','gaussian'}
                  kernel function defined as follows with

                  .. math::

                      z_{i,j} = d_{i,j}/h_i

                  triangular

                  .. math::

                      K(z) = (1 - |z|) \\ if |z| \\le 1

                  uniform

                  .. math::

                      K(z) = 1/2 \\ if |z| \\le 1

                  quadratic

                  .. math::

                      K(z) = (3/4)(1-z^2) \\ if |z| \\le 1

                  quartic

                  .. math::

                      K(z) = (15/16)(1-z^2)^2 \\ if |z| \\le 1

                  gaussian

                  .. math::

                      K(z) = (2\\pi)^{(-1/2)} exp(-z^2 / 2)

    eps         : float
                  adjustment to ensure knn distance range is closed on the
                  knnth observations

    Attributes
    ----------
    weights : dict
              Dictionary keyed by id with a list of weights for each neighbor

    neighbors : dict
                of lists of neighbors keyed by observation id

    bandwidth : array
                array of bandwidths

    Examples
    --------
    >>> from libpysal.weights import Kernel
    >>> points=[(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
    >>> kw=Kernel(points)
    >>> kw.weights[0]
    [1.0, 0.500000049999995, 0.4409830615267465]
    >>> kw.neighbors[0]
    [0, 1, 3]
    >>> kw.bandwidth
    array([[20.000002],
           [20.000002],
           [20.000002],
           [20.000002],
           [20.000002],
           [20.000002]])
    >>> kw15=Kernel(points,bandwidth=15.0)
    >>> kw15[0]
    {0: 1.0, 1: 0.33333333333333337, 3: 0.2546440075000701}
    >>> kw15.neighbors[0]
    [0, 1, 3]
    >>> kw15.bandwidth
    array([[15.],
           [15.],
           [15.],
           [15.],
           [15.],
           [15.]])

    Adaptive bandwidths user specified

    >>> bw=[25.0,15.0,25.0,16.0,14.5,25.0]
    >>> kwa=Kernel(points,bandwidth=bw)
    >>> kwa.weights[0]
    [1.0, 0.6, 0.552786404500042, 0.10557280900008403]
    >>> kwa.neighbors[0]
    [0, 1, 3, 4]
    >>> kwa.bandwidth
    array([[25. ],
           [15. ],
           [25. ],
           [16. ],
           [14.5],
           [25. ]])

    Endogenous adaptive bandwidths

    >>> kwea=Kernel(points,fixed=False)
    >>> kwea.weights[0]
    [1.0, 0.10557289844279438, 9.99999900663795e-08]
    >>> kwea.neighbors[0]
    [0, 1, 3]
    >>> kwea.bandwidth
    array([[11.18034101],
           [11.18034101],
           [20.000002  ],
           [11.18034101],
           [14.14213704],
           [18.02775818]])

    Endogenous adaptive bandwidths with Gaussian kernel

    >>> kweag=Kernel(points,fixed=False,function='gaussian')
    >>> kweag.weights[0]
    [0.3989422804014327, 0.2674190291577696, 0.2419707487162134]
    >>> kweag.bandwidth
    array([[11.18034101],
           [11.18034101],
           [20.000002  ],
           [11.18034101],
           [14.14213704],
           [18.02775818]])

    Diagonals to 1.0

    >>> kq = Kernel(points,function='gaussian')
    >>> kq.weights
    {0: [0.3989422804014327, 0.35206533556593145, 0.3412334260702758], 1: [0.35206533556593145, 0.3989422804014327, 0.2419707487162134, 0.3412334260702758, 0.31069657591175387], 2: [0.2419707487162134, 0.3989422804014327, 0.31069657591175387], 3: [0.3412334260702758, 0.3412334260702758, 0.3989422804014327, 0.3011374490937829, 0.26575287272131043], 4: [0.31069657591175387, 0.31069657591175387, 0.3011374490937829, 0.3989422804014327, 0.35206533556593145], 5: [0.26575287272131043, 0.35206533556593145, 0.3989422804014327]}
    >>> kqd = Kernel(points, function='gaussian', diagonal=True)
    >>> kqd.weights
    {0: [1.0, 0.35206533556593145, 0.3412334260702758], 1: [0.35206533556593145, 1.0, 0.2419707487162134, 0.3412334260702758, 0.31069657591175387], 2: [0.2419707487162134, 1.0, 0.31069657591175387], 3: [0.3412334260702758, 0.3412334260702758, 1.0, 0.3011374490937829, 0.26575287272131043], 4: [0.31069657591175387, 0.31069657591175387, 0.3011374490937829, 1.0, 0.35206533556593145], 5: [0.26575287272131043, 0.35206533556593145, 1.0]}

    """

    def __init__(
        self,
        data,
        bandwidth=None,
        fixed=True,
        k=2,
        function="triangular",
        eps=1.0000001,
        ids=None,
        diagonal=False,
        distance_metric="euclidean",
        radius=None,
        **kwargs
    ):
        if radius is not None:
            distance_metric = "arc"
        if isKDTree(data):
            self.kdtree = data
            self.data = self.kdtree.data
            data = self.data
        else:
            self.kdtree = KDTree(data, distance_metric=distance_metric, radius=radius)
            self.data = self.kdtree.data
        self.k = k + 1
        self.function = function.lower()
        self.fixed = fixed
        self.eps = eps
        if bandwidth:
            try:
                bandwidth = np.array(bandwidth)
                bandwidth.shape = (len(bandwidth), 1)
            except:
                bandwidth = np.ones((len(data), 1), "float") * bandwidth
            self.bandwidth = bandwidth
        else:
            self._set_bw()

        self._eval_kernel()
        neighbors, weights = self._k_to_W(ids)
        if diagonal:
            for i in neighbors:
                weights[i][neighbors[i].index(i)] = 1.0
        W.__init__(self, neighbors, weights, ids, **kwargs)

    @classmethod
    def from_shapefile(cls, filepath, idVariable=None, **kwargs):
        """
        Kernel based weights from shapefile

        Parameters
        ----------
        shapefile   : string
                      shapefile name with shp suffix
        idVariable  : string
                      name of column in shapefile's DBF to use for ids

        Returns
        -------
        Kernel Weights Object

        See Also
        ---------
        :class:`libpysal.weights.weights.W`
        """
        points = get_points_array_from_shapefile(filepath)
        if idVariable is not None:
            ids = get_ids(filepath, idVariable)
        else:
            ids = None
        return cls.from_array(points, ids=ids, **kwargs)

    @classmethod
    def from_array(cls, array, **kwargs):
        """
        Construct a Kernel weights from an array. Supports all the same options
        as :class:`libpysal.weights.Kernel`

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        return cls(array, **kwargs)

    @classmethod
    def from_dataframe(cls, df, geom_col=None, ids=None, **kwargs):
        """
        Make Kernel weights from a dataframe.

        Parameters
        ----------
        df      :   pandas.dataframe
                    a dataframe with a geometry column that can be used to
                    construct a W object
        geom_col :  string
                    the name of the column in `df` that contains the
                    geometries. Defaults to active geometry column.
        ids     :   string or iterable
                    if string, the column name of the indices from the dataframe
                    if iterable, a list of ids to use for the W
                    if None, df.index is used.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        if geom_col is None:
            geom_col = df.geometry.name
        pts = get_points_array(df[geom_col])
        if ids is None:
            ids = df.index.tolist()
        elif isinstance(ids, str):
            ids = df[ids].tolist()
        return cls(pts, ids=ids, **kwargs)

    def _k_to_W(self, ids=None):
        allneighbors = {}
        weights = {}
        if ids:
            ids = np.array(ids)
        else:
            ids = np.arange(len(self.data))
        for i, neighbors in enumerate(self.kernel):
            if len(self.neigh[i]) == 0:
                allneighbors[ids[i]] = []
                weights[ids[i]] = []
            else:
                allneighbors[ids[i]] = list(ids[self.neigh[i]])
                weights[ids[i]] = self.kernel[i].tolist()
        return allneighbors, weights

    def _set_bw(self):
        dmat, neigh = self.kdtree.query(self.data, k=self.k)
        if self.fixed:
            # use max knn distance as bandwidth
            bandwidth = dmat.max() * self.eps
            n = len(dmat)
            self.bandwidth = np.ones((n, 1), "float") * bandwidth
        else:
            # use local max knn distance
            self.bandwidth = dmat.max(axis=1) * self.eps
            self.bandwidth.shape = (self.bandwidth.size, 1)
            # identify knn neighbors for each point
            nnq = self.kdtree.query(self.data, k=self.k)
            self.neigh = nnq[1]

    def _eval_kernel(self):
        # get points within bandwidth distance of each point
        if not hasattr(self, "neigh"):
            kdtq = self.kdtree.query_ball_point
            neighbors = [
                kdtq(self.data[i], r=bwi[0]) for i, bwi in enumerate(self.bandwidth)
            ]
            self.neigh = neighbors
        # get distances for neighbors
        bw = self.bandwidth

        kdtq = self.kdtree.query
        z = []
        for i, nids in enumerate(self.neigh):
            di, ni = kdtq(self.data[i], k=len(nids))
            if not isinstance(di, np.ndarray):
                di = np.asarray([di] * len(nids))
                ni = np.asarray([ni] * len(nids))
            zi = np.array([dict(list(zip(ni, di)))[nid] for nid in nids]) / bw[i]
            z.append(zi)
        zs = z
        # functions follow Anselin and Rey (2010) table 5.4
        if self.function == "triangular":
            self.kernel = [1 - zi for zi in zs]
        elif self.function == "uniform":
            self.kernel = [np.ones(zi.shape) * 0.5 for zi in zs]
        elif self.function == "quadratic":
            self.kernel = [(3.0 / 4) * (1 - zi ** 2) for zi in zs]
        elif self.function == "quartic":
            self.kernel = [(15.0 / 16) * (1 - zi ** 2) ** 2 for zi in zs]
        elif self.function == "gaussian":
            c = np.pi * 2
            c = c ** (-0.5)
            self.kernel = [c * np.exp(-(zi ** 2) / 2.0) for zi in zs]
        else:
            print(("Unsupported kernel function", self.function))


class DistanceBand(W):
    """
    Spatial weights based on distance band.

    Parameters
    ----------

    data        : array
                  (n,k) or KDTree where KDtree.data is array (n,k)
                  n observations on k characteristics used to measure
                  distances between the n objects
    threshold  : float
                 distance band
    p          : float
                 DEPRECATED: use `distance_metric`
                 Minkowski p-norm distance metric parameter:
                 1<=p<=infinity
                 2: Euclidean distance
                 1: Manhattan distance
    binary     : boolean
                 If true w_{ij}=1 if d_{i,j}<=threshold, otherwise w_{i,j}=0
                 If false wij=dij^{alpha}
    alpha      : float
                 distance decay parameter for weight (default -1.0)
                 if alpha is positive the weights will not decline with
                 distance. If binary is True, alpha is ignored

    ids         : list
                  values to use for keys of the neighbors and weights dicts

    build_sp    : boolean
                  DEPRECATED
                  True to build sparse distance matrix and false to build dense
                  distance matrix; significant speed gains may be obtained
                  dending on the sparsity of the of distance_matrix and
                  threshold that is applied
    silent      : boolean
                  By default libpysal will print a warning if the
                  dataset contains any disconnected observations or
                  islands. To silence this warning set this
                  parameter to True.

    Attributes
    ----------
    weights : dict
              of neighbor weights keyed by observation id

    neighbors : dict
                of neighbors keyed by observation id

    Examples
    --------
    >>> import libpysal
    >>> points=[(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
    >>> wcheck = libpysal.weights.W({0: [1, 3], 1: [0, 3], 2: [], 3: [0, 1], 4: [5], 5: [4]})

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [2]
    >>> w=libpysal.weights.DistanceBand(points,threshold=11.2)

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [2]
    >>> libpysal.weights.util.neighbor_equality(w, wcheck)
    True
    >>> w=libpysal.weights.DistanceBand(points,threshold=14.2)
    >>> wcheck = libpysal.weights.W({0: [1, 3], 1: [0, 3, 4], 2: [4], 3: [1, 0], 4: [5, 2, 1], 5: [4]})
    >>> libpysal.weights.util.neighbor_equality(w, wcheck)
    True

    inverse distance weights

    >>> w=libpysal.weights.DistanceBand(points,threshold=11.2,binary=False)

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [2]
    >>> w.weights[0]
    [0.1, 0.08944271909999159]
    >>> w.neighbors[0].tolist()
    [1, 3]

    gravity weights

    >>> w=libpysal.weights.DistanceBand(points,threshold=11.2,binary=False,alpha=-2.)

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [2]
    >>> w.weights[0]
    [0.01, 0.007999999999999998]


    """

    def __init__(
        self,
        data,
        threshold,
        p=2,
        alpha=-1.0,
        binary=True,
        ids=None,
        build_sp=True,
        silence_warnings=False,
        distance_metric="euclidean",
        radius=None,
    ):
        """Casting to floats is a work around for a bug in scipy.spatial.
        See detail in pysal issue #126.

        """
        if ids is not None:
            ids = list(ids)
        if radius is not None:
            distance_metric = "arc"
        self.p = p
        self.threshold = threshold
        self.binary = binary
        self.alpha = alpha
        self.build_sp = build_sp
        self.silence_warnings = silence_warnings

        if isKDTree(data):
            self.kdtree = data
            self.data = self.kdtree.data
        else:
            if self.build_sp:
                try:
                    data = np.asarray(data)
                    if data.dtype.kind != "f":
                        data = data.astype(float)
                    self.kdtree = KDTree(
                        data, distance_metric=distance_metric, radius=radius
                    )
                    self.data = self.kdtree.data
                except:
                    raise ValueError("Could not make array from data")
            else:
                self.data = data
                self.kdtree = None

        self._band()
        neighbors, weights = self._distance_to_W(ids)
        W.__init__(
            self, neighbors, weights, ids, silence_warnings=self.silence_warnings
        )

    @classmethod
    def from_shapefile(cls, filepath, threshold, idVariable=None, **kwargs):
        """
        Distance-band based weights from shapefile

        Parameters
        ----------
        shapefile   : string
                      shapefile name with shp suffix
        idVariable  : string
                      name of column in shapefile's DBF to use for ids

        Returns
        --------
        Kernel Weights Object

        """
        points = get_points_array_from_shapefile(filepath)
        if idVariable is not None:
            ids = get_ids(filepath, idVariable)
        else:
            ids = None
        return cls.from_array(points, threshold, ids=ids, **kwargs)

    @classmethod
    def from_array(cls, array, threshold, **kwargs):
        """
        Construct a DistanceBand weights from an array. Supports all the same options
        as :class:`libpysal.weights.DistanceBand`

        """
        return cls(array, threshold, **kwargs)

    @classmethod
    def from_dataframe(cls, df, threshold, geom_col=None, ids=None, **kwargs):

        """
        Make DistanceBand weights from a dataframe.

        Parameters
        ----------
        df      :   pandas.dataframe
                    a dataframe with a geometry column that can be used to
                    construct a W object
        geom_col :  string
                    the name of the column in `df` that contains the
                    geometries. Defaults to active geometry column.
        ids     :   string or iterable
                    if string, the column name of the indices from the dataframe
                    if iterable, a list of ids to use for the W
                    if None, df.index is used.

        """
        if geom_col is None:
            geom_col = df.geometry.name
        pts = get_points_array(df[geom_col])
        if ids is None:
            ids = df.index.tolist()
        elif isinstance(ids, str):
            ids = df[ids].tolist()
        return cls(pts, threshold, ids=ids, **kwargs)

    def _band(self):
        """Find all pairs within threshold."""
        if self.build_sp:
            self.dmat = self.kdtree.sparse_distance_matrix(
                self.kdtree, max_distance=self.threshold, p=self.p
            ).tocsr()
        else:
            if str(self.kdtree).split(".")[-1][0:10] == "Arc_KDTree":
                raise TypeError(
                    "Unable to calculate dense arc distance matrix;"
                    ' parameter "build_sp" must be set to True for arc'
                    " distance type weight"
                )
            self.dmat = self._spdistance_matrix(self.data, self.data, self.threshold)

    def _distance_to_W(self, ids=None):
        if self.binary:
            self.dmat[self.dmat > 0] = 1
            self.dmat.eliminate_zeros()
            tempW = WSP2W(
                WSP(self.dmat, id_order=ids), silence_warnings=self.silence_warnings
            )
            neighbors = tempW.neighbors
            weight_keys = list(tempW.weights.keys())
            weight_vals = list(tempW.weights.values())
            weights = dict(list(zip(weight_keys, list(map(list, weight_vals)))))
            return neighbors, weights
        else:
            weighted = self.dmat.power(self.alpha)
            weighted[weighted == np.inf] = 0
            weighted.eliminate_zeros()
            tempW = WSP2W(
                WSP(weighted, id_order=ids), silence_warnings=self.silence_warnings
            )
            neighbors = tempW.neighbors
            weight_keys = list(tempW.weights.keys())
            weight_vals = list(tempW.weights.values())
            weights = dict(list(zip(weight_keys, list(map(list, weight_vals)))))
            return neighbors, weights

    def _spdistance_matrix(self, x, y, threshold=None):
        dist = distance_matrix(x, y)
        if threshold is not None:
            zeros = dist > threshold
            dist[zeros] = 0
        return sp.csr_matrix(dist)


def _test():
    import doctest

    # the following line could be used to define an alternative to the '<BLANKLINE>' flag
    # doctest.BLANKLINE_MARKER = 'something better than <BLANKLINE>'
    start_suppress = np.get_printoptions()["suppress"]
    np.set_printoptions(suppress=True)
    doctest.testmod()
    np.set_printoptions(suppress=start_suppress)


if __name__ == "__main__":
    _test()



"""Handle remote datasets."""

from bs4 import BeautifulSoup
import requests
import warnings
from .base import Example


def poll_remotes():
    """Fetch remote data and generate example datasets.

    Returns
    -------
    datasets : dict
        Example datasets keyed by the dataset name.

    """
    # Geoda Center Data Sets

    url = "https://geodacenter.github.io/data-and-lab//"
    try:
        page = requests.get(url)
    except:
        warnings.warn("Remote data sets not available. Check connection.")
        return {}
    soup = BeautifulSoup(page.text, "html.parser")
    samples = soup.find(class_="samples")
    rows = samples.find_all("tr")
    datasets = {}
    for row in rows[1:]:
        data = row.find_all("td")
        name = data[0].text.strip()
        description = data[1].text
        n = data[2].text
        k = data[3].text
        targets = row.find_all("a")
        download_url = url + targets[1].attrs["href"]
        explain_url = url + targets[0].attrs["href"]
        datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # Other Remotes

    # rio
    name = "Rio Grande do Sul"
    description = "Cities of the Brazilian State of Rio Grande do Sul"
    n = 497
    k = 3
    download_url = "https://github.com/sjsrey/rio_grande_do_sul/archive/master.zip"
    explain_url = (
        "https://raw.githubusercontent.com/sjsrey/rio_grande_do_sul/master/README.md"
    )
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # nyc bikes
    name = "nyc_bikes"
    description = "New York City Bike Trips"
    n = 14042
    k = 27
    download_url = "https://github.com/sjsrey/nyc_bikes/archive/master.zip"
    explain_url = "https://raw.githubusercontent.com/sjsrey/nyc_bikes/master/README.md"
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # taz
    name = "taz"
    description = "Traffic Analysis Zones in So. California"
    n = 4109
    k = 14
    download_url = "https://github.com/sjsrey/taz/archive/master.zip"
    explain_url = "https://raw.githubusercontent.com/sjsrey/taz/master/README.md"
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # clearwater
    name = "clearwater"
    description = "mgwr testing dataset"
    n = 239
    k = 14
    download_url = "https://github.com/sjsrey/clearwater/archive/master.zip"
    explain_url = "https://raw.githubusercontent.com/sjsrey/clearwater/master/README.md"
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # newHaven
    name = "newHaven"
    description = "Network testing dataset"
    n = 3293
    k = 5
    download_url = "https://github.com/sjsrey/newHaven/archive/master.zip"
    explain_url = "https://raw.githubusercontent.com/sjsrey/newHaven/master/README.md"
    datasets[name] = Example(name, description, n, k, download_url, explain_url)

    # remove Cars dataset as it is broken
    datasets.pop("Cars")

    return datasets


class Remotes:
    """Remote datasets."""

    def __init__(self):
        """Initialize Remotes."""
        self._datasets = None

    @property
    def datasets(self):
        """Create dictionary of remotes."""
        if self._datasets is None:
            self._datasets = poll_remotes()
        return self._datasets


datasets = Remotes()



# coding: utf-8

from setuptools import setup, find_packages
import versioneer

from distutils.command.build_py import build_py

import os

with open("README.rst", "r", encoding="utf8") as file:
    long_description = file.read()

# BEFORE importing distutils, remove MANIFEST. distutils doesn't properly
# update it when the contents of directories change.
if os.path.exists("MANIFEST"):
    os.remove("MANIFEST")


def _get_requirements_from_files(groups_files):
    groups_reqlist = {}

    for k, v in groups_files.items():
        with open(v, "r") as f:
            pkg_list = f.read().splitlines()
        groups_reqlist[k] = pkg_list

    return groups_reqlist


def setup_package():
    # get all file endings and copy whole file names without a file suffix
    # assumes nested directories are only down one level
    _groups_files = {
        "base": "requirements.txt",
        "plus_conda": "requirements_plus_conda.txt",
        "plus_pip": "requirements_plus_pip.txt",
        "dev": "requirements_dev.txt",
        "docs": "requirements_docs.txt",
    }

    reqs = _get_requirements_from_files(_groups_files)
    install_reqs = reqs.pop("base")
    extras_reqs = reqs

    # get all file endings and copy whole file names without a file suffix
    # assumes nested directories are only down one level
    example_data_files = set()
    for i in os.listdir("libpysal/examples"):
        if i.endswith(("py", "pyc")):
            continue
        if not os.path.isdir("libpysal/examples/" + i):
            if "." in i:
                glob_name = "examples/*." + i.split(".")[-1]
            else:
                glob_name = "examples/" + i
        else:
            glob_name = "examples/" + i + "/*"

        example_data_files.add(glob_name)

    setup(
        name="libpysal",
        version=versioneer.get_version(),
        description="Core components of PySAL A library of spatial analysis functions.",
        long_description=long_description,
        long_description_content_type="text/x-rst",
        maintainer="PySAL Developers",
        maintainer_email="pysal-dev@googlegroups.com",
        url="http://pysal.org/libpysal",
        download_url="https://pypi.python.org/pypi/libpysal",
        license="BSD",
        py_modules=["libpysal"],
        packages=find_packages(),
        setup_requires=["pytest-runner"],
        tests_require=["pytest"],
        keywords="spatial statistics",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: GIS",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
        ],
        package_data={"libpysal": list(example_data_files)},
        install_requires=install_reqs,
        extras_require=extras_reqs,
        cmdclass=versioneer.get_cmdclass({"build_py": build_py}),
        python_requires=">=3.7",
    )


if __name__ == "__main__":
    setup_package()



"""A Pure Python Shapefile Reader and Writer

This module is self-contained and does not require pysal. It returns and expects
dictionary-based data structures. It should be wrapped into your native data structures.

Contact:
    Charles Schmidt
    GeoDa Center
    Arizona State University
    Tempe, AZ
    http://geodacenter.asu.edu

"""

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"

from struct import calcsize, unpack, pack

from itertools import islice
import array
import sys
import io

from typing import Union


if sys.byteorder == "little":
    SYS_BYTE_ORDER = "<"
else:
    SYS_BYTE_ORDER = ">"

STRUCT_ITEMSIZE = {}
STRUCT_ITEMSIZE["i"] = calcsize("i")
STRUCT_ITEMSIZE["d"] = calcsize("d")

__all__ = ["shp_file", "shx_file"]

# SHAPEFILE Globals


def struct2arrayinfo(struct: tuple) -> list:
    struct = list(struct)
    names = [x[0] for x in struct]
    types = [x[1] for x in struct]
    orders = [x[2] for x in struct]
    lname, ltype, lorder = struct.pop(0)
    groups = {}
    g = 0
    groups[g] = {
        "names": [lname],
        "size": STRUCT_ITEMSIZE[ltype],
        "fmt": ltype,
        "order": lorder,
    }

    while struct:
        name, type, order = struct.pop(0)

        if order == lorder:
            groups[g]["names"].append(name)
            groups[g]["size"] += STRUCT_ITEMSIZE[type]
            groups[g]["fmt"] += type
        else:
            g += 1
            groups[g] = {
                "names": [name],
                "size": STRUCT_ITEMSIZE[type],
                "fmt": type,
                "order": order,
            }
        lname, ltype, lorder = name, type, order

    return [groups[x] for x in range(g + 1)]


HEADERSTRUCT = (
    ("File Code", "i", ">"),
    ("Unused0", "i", ">"),
    ("Unused1", "i", ">"),
    ("Unused2", "i", ">"),
    ("Unused3", "i", ">"),
    ("Unused4", "i", ">"),
    ("File Length", "i", ">"),
    ("Version", "i", "<"),
    ("Shape Type", "i", "<"),
    ("BBOX Xmin", "d", "<"),
    ("BBOX Ymin", "d", "<"),
    ("BBOX Xmax", "d", "<"),
    ("BBOX Ymax", "d", "<"),
    ("BBOX Zmin", "d", "<"),
    ("BBOX Zmax", "d", "<"),
    ("BBOX Mmin", "d", "<"),
    ("BBOX Mmax", "d", "<"),
)
UHEADERSTRUCT = struct2arrayinfo(HEADERSTRUCT)
RHEADERSTRUCT = (("Record Number", "i", ">"), ("Content Length", "i", ">"))
URHEADERSTRUCT = struct2arrayinfo(RHEADERSTRUCT)


def noneMax(a: Union[float, None], b: Union[float, None]) -> float:
    if a is None:
        return b
    if b is None:
        return a
    return max(a, b)


def noneMin(a: Union[float, None], b: Union[float, None]) -> float:
    if a is None:
        return b
    if b is None:
        return a
    return min(a, b)


def _unpackDict(structure, fileObj):
    """Utility function that requires a tuple of tuples that
    describe the element structure.
    
    Parameters
    ----------
    structure : tuple
        A tuple of tuples in the form:
        ``(('FieldName 1','type','byteOrder'),('FieldName 2','type','byteOrder'))``.
    fileObj : file
        An open file at the correct position.
    
    Returns
    -------
    d : dict
        Dictionary in the form: ``{'FieldName 1': value, 'FieldName 2': value}``.
    
    Notes
    -----
    
    The file is at new position.
    
    Examples
    --------
    
    >>> import libpysal
    >>> _unpackDict(
    ...     UHEADERSTRUCT,
    ...     open(
    ...         libpysal.examples.get_path('10740.shx'), 'rb')
    ... ) == \
    ... {
    ...     'BBOX Xmax': -105.29012,
    ...     'BBOX Ymax': 36.219799000000002,
    ...     'BBOX Mmax': 0.0,
    ...     'BBOX Zmin': 0.0,
    ...     'BBOX Mmin': 0.0,
    ...     'File Code': 9994,
    ...     'BBOX Ymin': 34.259672000000002,
    ...     'BBOX Xmin': -107.62651,
    ...     'Unused0': 0,
    ...     'Unused1': 0,
    ...     'Unused2': 0,
    ...     'Unused3': 0,
    ...     'Unused4': 0,
    ...     'Version': 1000,
    ...     'BBOX Zmax': 0.0,
    ...     'Shape Type': 5,
    ...     'File Length': 830
    ... }
    True
    
    """

    d = {}

    for struct in structure:
        items = unpack(struct["order"] + struct["fmt"], fileObj.read(struct["size"]))
        for i, name in enumerate(struct["names"]):
            d[name] = items[i]

    return d


def _unpackDict2(d, structure, fileObj):
    """Utility Function, used arrays instead from struct.
    
    Parameters
    ----------
    d : dict
        Dictionary in to be updated.
    structure : tuple
        A tuple of tuples in the form:
        ``(('FieldName 1','type','byteOrder'),('FieldName 2','type','byteOrder'))``.
    fileObj : file
        An open file at the correct position.
    
    Returns
    -------
    d : dict
        The updated dictionary.
    
    """

    for name, dtype, order in structure:
        dtype, n = dtype
        result = array.array(dtype)
        result.frombytes(fileObj.read(result.itemsize * n))
        if order != SYS_BYTE_ORDER:
            result.byteswap()
        d[name] = result.tolist()

    return d


def _packDict(structure, d) -> str:
    """Utility Function for packing a dictionary with byte strings.
    
    Parameters
    ----------
    structure : tuple
        A tuple of tuples in the form:
        ``(('FieldName 1','type','byteOrder'),('FieldName 2','type','byteOrder'))``.
    d : dict
        Dictionary in the form: ``{'FieldName 1': value, 'FieldName 2': value}``.
    
    Examples
    --------
    
    >>> s = _packDict(
    ...     (('FieldName 1', 'i', '<'), ('FieldName 2', 'i', '<')),
    ...     {'FieldName 1': 1, 'FieldName 2': 2}
    ... )
    >>> s == pack('<ii', 1, 2)
    True
    
    >>> unpack('<ii', s)
    (1, 2)
    
    """

    string = b""

    for name, dtype, order in structure:
        if len(dtype) > 1:
            string += pack(order + dtype, *d[name])
        else:
            string += pack(order + dtype, d[name])

    return string


class shp_file:
    """Reads and writes the SHP compenent of a shapefile.
    
    Parameters
    ----------
    filename : str
        The name of the file to create.
    mode : str
        The mode for file interaction, either ``'r'`` (read)
        or ``'w'`` (write). Default is ``'r'``.
    shape_type : str
        Must be one of the following: ``'POINT'``, ``'POINTZ'``, ``'POINTM'``,
        ``'ARC'``, ``'ARCZ'``, ``'ARCM'``, ``'POLYGON'``, ``'POLYGONZ'``,
        ``'POLYGONM'``, ``'MULTIPOINT'``, ``'MULTIPOINTZ'``, ``'MULTIPOINTM'``,
        ``'MULTIPATCH'``. Default is ``None``.
    
    Attributes
    ----------
    header : dict
        Contents of the SHP header. For contents see ``HEADERSTRUCT``.
    shape : int
        See ``SHAPE_TYPES`` and ``TYPE_DISPATCH``.
    
    Examples
    --------
    
    >>> import libpysal
    >>> shp = shp_file(libpysal.examples.get_path('10740.shp'))
    >>> shp.header == {
    ...     'BBOX Xmax': -105.29012,
    ...     'BBOX Ymax': 36.219799000000002,
    ...     'BBOX Mmax': 0.0,
    ...     'BBOX Zmin': 0.0,
    ...     'BBOX Mmin': 0.0,
    ...     'File Code': 9994,
    ...     'BBOX Ymin': 34.259672000000002,
    ...     'BBOX Xmin': -107.62651,
    ...     'Unused0': 0,
    ...     'Unused1': 0,
    ...     'Unused2': 0,
    ...     'Unused3': 0,
    ...     'Unused4': 0,
    ...     'Version': 1000,
    ...     'BBOX Zmax': 0.0,
    ...     'Shape Type': 5,
    ...     'File Length': 260534
    ... }
    True
    
    >>> len(shp)
    195
    
    Notes
    -----
    
    The header of both the SHP and SHX files are indentical.
    
    """

    SHAPE_TYPES = {
        "POINT": 1,
        "ARC": 3,
        "POLYGON": 5,
        "MULTIPOINT": 8,
        "POINTZ": 11,
        "ARCZ": 13,
        "POLYGONZ": 15,
        "MULTIPOINTZ": 18,
        "POINTM": 21,
        "ARCM": 23,
        "POLYGONM": 25,
        "MULTIPOINTM": 28,
        "MULTIPATCH": 31,
    }

    def __iswritable(self) -> bool:
        """
        
        Raises
        ------
        IOError
            Raised when a bad file name is passed in.
        
        """

        try:
            assert self.__mode == "w"
        except AssertionError:
            raise IOError("[Errno 9] Bad file descriptor.")
        return True

    def __isreadable(self) -> bool:
        """
        
        Raises
        ------
        IOError
            Raised when a bad file name is passed in.
        
        """

        try:
            assert self.__mode == "r"
        except AssertionError:
            raise IOError("[Errno 9] Bad file descriptor.")
        return True

    def __init__(self, fileName, mode="r", shape_type=None):
        """
        
        Raises
        ------
        Exception
            Raised when an invalid shape type is passed in.
        Exception
            Raised when an invalid mode is passed in.
        
        """

        self.__mode = mode

        if (
            fileName.lower().endswith(".shp")
            or fileName.lower().endswith(".shx")
            or fileName.lower().endswith(".dbf")
        ):
            fileName = fileName[:-4]

        self.fileName = fileName

        if mode == "r":
            self._open_shp_file()
        elif mode == "w":
            if shape_type not in self.SHAPE_TYPES:
                raise Exception("Attempt to create shp/shx file of invalid type.")
            self._create_shp_file(shape_type)
        else:
            raise Exception("Only 'w' and 'r' modes are supported.")

    def _open_shp_file(self):
        """Opens a shp/shx file."""

        self.__isreadable()
        fileName = self.fileName
        self.fileObj = open(fileName + ".shp", "rb")
        self._shx = shx_file(fileName)
        self.header = _unpackDict(UHEADERSTRUCT, self.fileObj)
        self.shape = TYPE_DISPATCH[self.header["Shape Type"]]
        self.__lastShape = 0

        # localizing for convenience
        self.__numRecords = self._shx.numRecords

        # constructing bounding box from header
        h = self.header
        self.bbox = [h["BBOX Xmin"], h["BBOX Ymin"], h["BBOX Xmax"], h["BBOX Ymax"]]
        self.shapeType = self.header["Shape Type"]

    def _create_shp_file(self, shape_type: str):
        """Creates a shp/shx file.
        
        Examples
        --------
        
        >>> import libpysal, os
        >>> shp = shp_file('test', 'w', 'POINT')
        >>> p = shp_file(libpysal.examples.get_path('Point.shp'))
        >>> for pt in p:
        ...   shp.add_shape(pt)
        >>> shp.close()
        >>> open('test.shp','rb').read() == open(
        ...     libpysal.examples.get_path('Point.shp'), 'rb'
        ... ).read()
        True
        
        >>> open('test.shx', 'rb').read() == open(
        ...     libpysal.examples.get_path('Point.shx'), 'rb'
        ... ).read()
        True
        
        >>> os.remove('test.shx')
        >>> os.remove('test.shp')
        
        """

        self.__iswritable()
        fileName = self.fileName
        self.fileObj = open(fileName + ".shp", "wb")
        self._shx = shx_file(fileName, "w")
        self.header = {}
        self.header["Shape Type"] = self.SHAPE_TYPES[shape_type]
        self.header["Version"] = 1000
        self.header["Unused0"] = 0
        self.header["Unused1"] = 0
        self.header["Unused2"] = 0
        self.header["Unused3"] = 0
        self.header["Unused4"] = 0
        self.header["File Code"] = 9994
        self.__file_Length = 100
        self.header["File Length"] = 0
        self.header["BBOX Xmax"] = None
        self.header["BBOX Ymax"] = None
        self.header["BBOX Mmax"] = None
        self.header["BBOX Zmax"] = None
        self.header["BBOX Xmin"] = None
        self.header["BBOX Ymin"] = None
        self.header["BBOX Mmin"] = None
        self.header["BBOX Zmin"] = None
        self.shape = TYPE_DISPATCH[self.header["Shape Type"]]
        # self.__numRecords = self._shx.numRecords

    def __len__(self) -> int:
        return self.__numRecords

    def __iter__(self):
        return self

    def type(self) -> str:
        return self.shape.String_Type

    def __next__(self) -> int:
        """Returns the next shape in the shapefile.
        
        Raises
        ------
        StopIteration
            Raised at the EOF.
        
        Examples
        --------
        
        >>> import libpysal
        >>> list(shp_file(libpysal.examples.get_path('Point.shp'))) == [
        ...     {
        ...         'Y': -0.25904661905760773,
        ...         'X': -0.00068176617532103578,
        ...         'Shape Type': 1
        ...     },
        ...     {
        ...         'Y': -0.25630328607387354,
        ...         'X': 0.11697145363360706,
        ...         'Shape Type': 1
        ...     },
        ...     {
        ...         'Y': -0.33930131004366804,
        ...         'X': 0.05043668122270728,
        ...         'Shape Type': 1
        ...     },
        ...     {
        ...         'Y': -0.41266375545851519,
        ...         'X': -0.041266375545851552,
        ...         'Shape Type': 1
        ...     },
        ...     {
        ...         'Y': -0.44017467248908293,
        ...         'X': -0.011462882096069604,
        ...         'Shape Type': 1
        ...     },
        ...     {
        ...         'Y': -0.46080786026200882,
        ...         'X': 0.027510917030567628,
        ...         'Shape Type': 1
        ...     },
        ...     {
        ...         'Y': -0.45851528384279472,
        ...         'X': 0.075655021834060809,
        ...         'Shape Type': 1
        ...     },
        ...     {
        ...         'Y': -0.43558951965065495,
        ...         'X': 0.11233624454148461,
        ...         'Shape Type': 1
        ...     },
        ...     {
        ...         'Y': -0.40578602620087334,
        ...         'X': 0.13984716157205224,
        ...         'Shape Type': 1
        ...     }
        ... ]
        True
        
        """

        self.__isreadable()
        nextShape = self.__lastShape
        if nextShape == self._shx.numRecords:
            self.__lastShape = 0
            raise StopIteration
        else:
            self.__lastShape = nextShape + 1
            return self.get_shape(nextShape)

    def __seek(self, pos: int):
        if pos != self.fileObj.tell():
            self.fileObj.seek(pos)

    def __read(self, pos: int, size: int):
        self.__isreadable()
        if pos != self.fileObj.tell():
            self.fileObj.seek(pos)
        return self.fileObj.read(size)

    def get_shape(self, shpId: int) -> dict:
        self.__isreadable()
        if shpId + 1 > self.__numRecords:
            raise IndexError
        fPosition, byts = self._shx.index[shpId]
        self.__seek(fPosition)

        # the index does not include the 2 byte record header
        # (which contains, Record ID and Content Length)

        rec_id, con_len = _unpackDict(URHEADERSTRUCT, self.fileObj)

        return self.shape.unpack(io.BytesIO(self.fileObj.read(byts)))
        # return self.shape.unpack(self.fileObj.read(bytes))

    def __update_bbox(self, s: dict):
        h = self.header
        if s.get("Shape Type") == 1:
            h["BBOX Xmax"] = noneMax(h["BBOX Xmax"], s.get("X"))
            h["BBOX Ymax"] = noneMax(h["BBOX Ymax"], s.get("Y"))
            h["BBOX Mmax"] = noneMax(h["BBOX Mmax"], s.get("M"))
            h["BBOX Zmax"] = noneMax(h["BBOX Zmax"], s.get("Z"))
            h["BBOX Xmin"] = noneMin(h["BBOX Xmin"], s.get("X"))
            h["BBOX Ymin"] = noneMin(h["BBOX Ymin"], s.get("Y"))
            h["BBOX Mmin"] = noneMin(h["BBOX Mmin"], s.get("M"))
            h["BBOX Zmin"] = noneMin(h["BBOX Zmin"], s.get("Z"))
        else:
            h["BBOX Xmax"] = noneMax(h["BBOX Xmax"], s.get("BBOX Xmax"))
            h["BBOX Ymax"] = noneMax(h["BBOX Ymax"], s.get("BBOX Ymax"))
            h["BBOX Mmax"] = noneMax(h["BBOX Mmax"], s.get("BBOX Mmax"))
            h["BBOX Zmax"] = noneMax(h["BBOX Zmax"], s.get("BBOX Zmax"))
            h["BBOX Xmin"] = noneMin(h["BBOX Xmin"], s.get("BBOX Xmin"))
            h["BBOX Ymin"] = noneMin(h["BBOX Ymin"], s.get("BBOX Ymin"))
            h["BBOX Mmin"] = noneMin(h["BBOX Mmin"], s.get("BBOX Mmin"))
            h["BBOX Zmin"] = noneMin(h["BBOX Zmin"], s.get("BBOX Zmin"))

        if not self.shape.HASM:
            self.header["BBOX Mmax"] = 0.0
            self.header["BBOX Mmin"] = 0.0
        if not self.shape.HASZ:
            self.header["BBOX Zmax"] = 0.0
            self.header["BBOX Zmin"] = 0.0

    def add_shape(self, s: dict):
        self.__iswritable()
        self.__update_bbox(s)
        rec = self.shape.pack(s)
        con_len = len(rec)
        self.__file_Length += con_len + 8
        rec_id, pos = self._shx.add_record(con_len)
        self.__seek(pos)
        self.fileObj.write(pack(">ii", rec_id, con_len // 2))
        self.fileObj.write(rec)

    def close(self):
        self._shx.close(self.header)
        if self.__mode == "w":
            self.header["File Length"] = self.__file_Length // 2
            self.__seek(0)
            self.fileObj.write(_packDict(HEADERSTRUCT, self.header))
        self.fileObj.close()


class shx_file:
    """Reads and writes the SHX compenent of a shapefile.
    
    Parameters
    ----------
    filename : str
        The name of the file to create. Default is ``None``.
        The extension is optional, will remove ``'.dbf'``,
        ``'.shx'``, ``'.shp'`` and append ``'.shx'``.
    mode : str
        The mode for file interaction. Must be ``'r'`` (read).
    
    Attributes
    ----------
    index : list
        Contains the file offset and length of each recond in the SHP component.
    numRecords : int
        The number of records.
    
    Examples
    --------
    
    >>> import libpysal
    >>> shx = shx_file(libpysal.examples.get_path('10740.shx'))
    >>> shx._header == {
    ...     'BBOX Xmax': -105.29012,
    ...     'BBOX Ymax': 36.219799000000002,
    ...     'BBOX Mmax': 0.0,
    ...     'BBOX Zmin': 0.0,
    ...     'BBOX Mmin': 0.0,
    ...     'File Code': 9994,
    ...     'BBOX Ymin': 34.259672000000002,
    ...     'BBOX Xmin': -107.62651,
    ...     'Unused0': 0,
    ...     'Unused1': 0,
    ...     'Unused2': 0,
    ...     'Unused3': 0,
    ...     'Unused4': 0,
    ...     'Version': 1000,
    ...     'BBOX Zmax': 0.0,
    ...     'Shape Type': 5,
    ...     'File Length': 830
    ... }
    True
    
    >>> len(shx.index)
    195
    
    >>> shx = shx_file(libpysal.examples.get_path('Point.shx'))
    >>> isinstance(shx, shx_file)
    True
    
    """

    def __iswritable(self) -> bool:
        """
        
        Raises
        ------
        IOError
            Raised when a bad file name is passed in.
        
        """

        try:
            assert self.__mode == "w"
        except AssertionError:
            raise IOError("[Errno 9] Bad file descriptor.")
        return True

    def __isreadable(self) -> bool:
        """
        
        Raises
        ------
        IOError
            Raised when a bad file name is passed in.
        
        """

        try:
            assert self.__mode == "r"
        except AssertionError:
            raise IOError("[Errno 9] Bad file descriptor.")
        return True

    def __init__(self, fileName=None, mode="r"):
        self.__mode = mode
        if (
            fileName.endswith(".shp")
            or fileName.endswith(".shx")
            or fileName.endswith(".dbf")
        ):
            fileName = fileName[:-4]
        self.fileName = fileName

        if mode == "r":
            self._open_shx_file()
        elif mode == "w":
            self._create_shx_file()

    def _open_shx_file(self):
        """Opens the SHX file."""

        self.__isreadable()
        self.fileObj = open(self.fileName + ".shx", "rb")
        self._header = _unpackDict(UHEADERSTRUCT, self.fileObj)
        self.numRecords = numRecords = (self._header["File Length"] - 50) // 4
        index = {}
        fmt = ">%di" % (2 * numRecords)
        size = calcsize(fmt)
        dat = unpack(fmt, self.fileObj.read(size))
        self.index = [(dat[i] * 2, dat[i + 1] * 2) for i in range(0, len(dat), 2)]

    def _create_shx_file(self):
        """Creates the SHX file."""

        self.__iswritable()
        self.fileObj = open(self.fileName + ".shx", "wb")
        self.numRecords = 0
        self.index = []

        # length of header
        self.__offset = 100
        # record IDs start at 1
        self.__next_rid = 1

    def add_record(self, size: int):
        """Add a record to the shx index.
        
        Parameters
        ----------
        size : int
            The length of the record in bytes NOT including the 8-byte record header.
        
        Returns
        -------
        rec_id : int
            The sequential record ID, 1-based.
        pos : int
            See ``self.__offset`` in ``_create_shx_file``.
        
        Notes
        -----
        
        The SHX records contain (Offset, Length) in 16-bit words.
        
        Examples
        --------
        
        >>> import libpysal, os
        >>> shx = shx_file(libpysal.examples.get_path('Point.shx'))
        >>> shx.index
        [(100, 20),
         (128, 20),
         (156, 20),
         (184, 20),
         (212, 20),
         (240, 20),
         (268, 20),
         (296, 20),
         (324, 20)]
        
        >>> shx2 = shx_file('test', 'w')
        >>> [shx2.add_record(rec[1]) for rec in shx.index]
        [(1, 100),
         (2, 128),
         (3, 156),
         (4, 184),
         (5, 212),
         (6, 240),
         (7, 268),
         (8, 296),
         (9, 324)]
        
        >>> shx2.index == shx.index
        True
        
        >>> shx2.close(shx._header)
        >>> open('test.shx', 'rb').read() == open(
        ...     libpysal.examples.get_path('Point.shx'), 'rb'
        ... ).read()
        True
        
        >>> os.remove('test.shx')
        
        """

        self.__iswritable()
        pos = self.__offset
        rec_id = self.__next_rid
        self.index.append((self.__offset, size))

        # the 8-byte record header.
        self.__offset += size + 8
        self.numRecords += 1
        self.__next_rid += 1

        return rec_id, pos

    def close(self, header: dict):
        if self.__mode == "w":
            self.__iswritable()
            header["File Length"] = (self.numRecords * calcsize(">ii") + 100) // 2
            self.fileObj.seek(0)
            self.fileObj.write(_packDict(HEADERSTRUCT, header))
            fmt = ">%di" % (2 * self.numRecords)
            values = []
            for off, size in self.index:
                values.extend([off // 2, size // 2])
            self.fileObj.write(pack(fmt, *values))

        self.fileObj.close()


class NullShape:
    Shape_Type = 0
    STRUCT = ("Shape Type", "i", "<")

    def unpack(self) -> None:
        return None

    def pack(self, x=None) -> str:
        return pack("<i", 0)


class Point(object):
    """Packs and unpacks a shapefile Point type.
    
    Examples
    --------
    
    >>> import libpysal
    >>> shp = shp_file(libpysal.examples.get_path('Point.shp'))
    >>> rec = shp.get_shape(0)
    >>> rec == {'Y': -0.25904661905760773, 'X': -0.00068176617532103578, 'Shape Type': 1}
    True
    
    >>> # +8 byte record header
    >>> pos = shp.fileObj.seek(shp._shx.index[0][0] + 8)
    >>> dat = shp.fileObj.read(shp._shx.index[0][1])
    >>> dat == Point.pack(rec)
    True
    
    """

    Shape_Type = 1
    String_Type = "POINT"
    HASZ = False
    HASM = False
    STRUCT = (("Shape Type", "i", "<"), ("X", "d", "<"), ("Y", "d", "<"))
    USTRUCT = [
        {"fmt": "idd", "order": "<", "names": ["Shape Type", "X", "Y"], "size": 20}
    ]

    @classmethod
    def unpack(cls, dat) -> dict:
        """
        Parameters
        ----------
        dat : file
            An open file at the correct position.
        """

        return _unpackDict(cls.USTRUCT, dat)

    @classmethod
    def pack(cls, record: dict) -> str:
        rheader = _packDict(cls.STRUCT, record)
        return rheader


class PointZ(Point):
    Shape_Type = 11
    String_Type = "POINTZ"
    HASZ = True
    HASM = True
    STRUCT = (
        ("Shape Type", "i", "<"),
        ("X", "d", "<"),
        ("Y", "d", "<"),
        ("Z", "d", "<"),
        ("M", "d", "<"),
    )
    USTRUCT = [
        {
            "fmt": "idddd",
            "order": "<",
            "names": ["Shape Type", "X", "Y", "Z", "M"],
            "size": 36,
        }
    ]


class PolyLine:
    """Packs and unpacks a shapefile PolyLine type.
    
    Examples
    --------
    
    >>> import libpysal
    >>> shp = shp_file(libpysal.examples.get_path('Line.shp'))
    >>> rec = shp.get_shape(0)
    >>> rec == {
    ...     'BBOX Ymax': -0.25832280562918325,
    ...     'NumPoints': 3,
    ...     'BBOX Ymin': -0.25895877033237352,
    ...     'NumParts': 1,
    ...     'Vertices': [
    ...         (-0.0090539248870159517, -0.25832280562918325),
    ...         (0.0074811573959305822, -0.25895877033237352),
    ...         (0.0074811573959305822, -0.25895877033237352)
    ...     ],
    ...     'BBOX Xmax': 0.0074811573959305822,
    ...     'BBOX Xmin': -0.0090539248870159517,
    ...     'Shape Type': 3,
    ...     'Parts Index': [0]
    ... }
    True
    
    >>> # +8 byte record header
    >>> pos = shp.fileObj.seek(shp._shx.index[0][0] + 8)
    >>> dat = shp.fileObj.read(shp._shx.index[0][1])
    >>> dat == PolyLine.pack(rec)
    True
    
    """

    HASZ = False
    HASM = False
    String_Type = "ARC"
    STRUCT = (
        ("Shape Type", "i", "<"),
        ("BBOX Xmin", "d", "<"),
        ("BBOX Ymin", "d", "<"),
        ("BBOX Xmax", "d", "<"),
        ("BBOX Ymax", "d", "<"),
        ("NumParts", "i", "<"),
        ("NumPoints", "i", "<"),
    )
    USTRUCT = [
        {
            "fmt": "iddddii",
            "order": "<",
            "names": [
                "Shape Type",
                "BBOX Xmin",
                "BBOX Ymin",
                "BBOX Xmax",
                "BBOX Ymax",
                "NumParts",
                "NumPoints",
            ],
            "size": 44,
        }
    ]

    @classmethod
    def unpack(cls, dat) -> dict:
        """
        
        Parameters
        ----------
        dat : file
            An open file at the correct position.
        
        """

        record = _unpackDict(cls.USTRUCT, dat)
        contentStruct = (
            ("Parts Index", ("i", record["NumParts"]), "<"),
            ("Vertices", ("d", 2 * record["NumPoints"]), "<"),
        )
        _unpackDict2(record, contentStruct, dat)

        # record['Vertices'] = [
        #    (record['Vertices'][i], record['Vertices'][i+1])
        #    for i in range(0, record['NumPoints']*2, 2)
        # ]
        verts = record["Vertices"]

        # Next line is equivalent to: zip(verts[::2],verts[1::2])
        record["Vertices"] = list(
            zip(islice(verts, 0, None, 2), islice(verts, 1, None, 2))
        )
        if not record["Parts Index"]:
            record["Parts Index"] = [0]

        return record
        # partsIndex = list(partsIndex)
        # partsIndex.append(None)
        # parts = [
        #     vertices[partsIndex[i]:partsIndex[i+1]] for i in range(header['NumParts'])
        # ]

    @classmethod
    def pack(cls, record: dict) -> str:
        rheader = _packDict(cls.STRUCT, record)
        contentStruct = (
            ("Parts Index", "%di" % record["NumParts"], "<"),
            ("Vertices", "%dd" % (2 * record["NumPoints"]), "<"),
        )
        content = {}
        content["Parts Index"] = record["Parts Index"]
        verts = []
        [verts.extend(vert) for vert in record["Vertices"]]
        content["Vertices"] = verts
        content = _packDict(contentStruct, content)

        return rheader + content


class PolyLineZ(object):
    HASZ = True
    HASM = True
    String_Type = "ARC"
    STRUCT = (
        ("Shape Type", "i", "<"),
        ("BBOX Xmin", "d", "<"),
        ("BBOX Ymin", "d", "<"),
        ("BBOX Xmax", "d", "<"),
        ("BBOX Ymax", "d", "<"),
        ("NumParts", "i", "<"),
        ("NumPoints", "i", "<"),
    )
    USTRUCT = [
        {
            "fmt": "iddddii",
            "order": "<",
            "names": [
                "Shape Type",
                "BBOX Xmin",
                "BBOX Ymin",
                "BBOX Xmax",
                "BBOX Ymax",
                "NumParts",
                "NumPoints",
            ],
            "size": 44,
        }
    ]

    @classmethod
    def unpack(cls, dat) -> dict:
        """
        
        Parameters
        ----------
        dat : file
            An open file at the correct position.
        
        """

        record = _unpackDict(cls.USTRUCT, dat)
        contentStruct = (
            ("Parts Index", ("i", record["NumParts"]), "<"),
            ("Vertices", ("d", 2 * record["NumPoints"]), "<"),
            ("Zmin", ("d", 1), "<"),
            ("Zmax", ("d", 1), "<"),
            ("Zarray", ("d", record["NumPoints"]), "<"),
            ("Mmin", ("d", 1), "<"),
            ("Mmax", ("d", 1), "<"),
            ("Marray", ("d", record["NumPoints"]), "<"),
        )

        _unpackDict2(record, contentStruct, dat)
        verts = record["Vertices"]
        record["Vertices"] = list(
            zip(islice(verts, 0, None, 2), islice(verts, 1, None, 2))
        )

        if not record["Parts Index"]:
            record["Parts Index"] = [0]
        record["Zmin"] = record["Zmin"][0]
        record["Zmax"] = record["Zmax"][0]
        record["Mmin"] = record["Mmin"][0]
        record["Mmax"] = record["Mmax"][0]

        return record

    @classmethod
    def pack(cls, record: dict) -> str:
        rheader = _packDict(cls.STRUCT, record)
        contentStruct = (
            ("Parts Index", "%di" % record["NumParts"], "<"),
            ("Vertices", "%dd" % (2 * record["NumPoints"]), "<"),
            ("Zmin", "d", "<"),
            ("Zmax", "d", "<"),
            ("Zarray", "%dd" % (record["NumPoints"]), "<"),
            ("Mmin", "d", "<"),
            ("Mmax", "d", "<"),
            ("Marray", "%dd" % (record["NumPoints"]), "<"),
        )

        content = {}
        content.update(record)
        content["Parts Index"] = record["Parts Index"]
        verts = []
        [verts.extend(vert) for vert in record["Vertices"]]
        content["Vertices"] = verts
        content = _packDict(contentStruct, content)

        return rheader + content


class Polygon(PolyLine):
    """Packs and unpacks a shapefile Polygon type identical to PolyLine.
    
    Examples
    --------
    
    >>> import libpysal
    >>> shp = shp_file(libpysal.examples.get_path('Polygon.shp'))
    >>> rec = shp.get_shape(1)
    >>> rec == {
    ...     'BBOX Ymax': -0.3126531125455273,
    ...     'NumPoints': 7,
    ...     'BBOX Ymin': -0.35957259110238166,
    ...     'NumParts': 1,
    ...     'Vertices': [
    ...         (0.05396439570183631, -0.3126531125455273),
    ...         (0.051473095955454629, -0.35251390848763364),
    ...         (0.059777428443393454, -0.34254870950210703),
    ...         (0.063099161438568974, -0.34462479262409174),
    ...         (0.048981796209073003, -0.35957259110238166),
    ...         (0.046905713087088297, -0.3126531125455273),
    ...         (0.05396439570183631, -0.3126531125455273)
    ...     ],
    ...     'BBOX Xmax': 0.063099161438568974,
    ...     'BBOX Xmin': 0.046905713087088297,
    ...     'Shape Type': 5,
    ...     'Parts Index': [0]
    ...     }
    True
    
    >>> # +8 byte record header
    >>> pos = shp.fileObj.seek(shp._shx.index[1][0] + 8)
    >>> dat = shp.fileObj.read(shp._shx.index[1][1])
    >>> dat == Polygon.pack(rec)
    True
    
    """

    String_Type = "POLYGON"


class MultiPoint:
    String_Type = "MULTIPOINT"

    def __init__(self):
        raise NotImplementedError("No MultiPoint support at this time.")


class PolygonZ(PolyLineZ):
    String_Type = "POLYGONZ"


class MultiPointZ:
    String_Type = "MULTIPOINTZ"

    def __init__(self):
        raise NotImplementedError("No MultiPointZ support at this time.")


class PointM:
    String_Type = "POINTM"

    def __init__(self):
        raise NotImplementedError("No PointM support at this time.")


class PolyLineM:
    String_Type = "ARCM"

    def __init__(self):
        raise NotImplementedError("No PolyLineM support at this time.")


class PolygonM:
    String_Type = "POLYGONM"

    def __init__(self):
        raise NotImplementedError("No PolygonM support at this time.")


class MultiPointM:
    String_Type = "MULTIPOINTM"

    def __init__(self):
        raise NotImplementedError("No MultiPointM support at this time.")


class MultiPatch:
    String_Type = "MULTIPATCH"

    def __init__(self):
        raise NotImplementedError("No MultiPatch support at this time.")


TYPE_DISPATCH = {
    0: NullShape,
    1: Point,
    3: PolyLine,
    5: Polygon,
    8: MultiPoint,
    11: PointZ,
    13: PolyLineZ,
    15: PolygonZ,
    18: MultiPointZ,
    21: PointM,
    23: PolyLineM,
    25: PolygonM,
    28: MultiPointM,
    31: MultiPatch,
    "POINT": Point,
    "POINTZ": PointZ,
    "POINTM": PointM,
    "ARC": PolyLine,
    "ARCZ": PolyLineZ,
    "ARCM": PolyLineM,
    "POLYGON": Polygon,
    "POLYGONZ": PolygonZ,
    "POLYGONM": PolygonM,
    "MULTIPOINT": MultiPoint,
    "MULTIPOINTZ": MultiPointZ,
    "MULTIPOINTM": MultiPointM,
    "MULTIPATCH": MultiPatch,
}
